//! Extension modules for seal bindings.

/// Operations in batch of ciphertexts. It allows us to perform operations
/// on multiple ciphertexts at once, dribbling the size limits of the scheme.
pub mod batched;
